# queue.py

import unittest
from collections import deque

from ..import *

# ---- Queuing

class TestQueue(unittest.TestCase):
    def test_queue_simple(self):
        kernel = get_kernel()
        results = []
        async def consumer(queue, label):
              while True:
                  item = await queue.get()
                  if item is None:
                      break
                  results.append((label, item))
                  await queue.task_done()
              await queue.task_done()
              results.append(label + ' done')

        async def producer():
            queue = Queue()
            results.append('producer_start')
            await new_task(consumer(queue, 'cons1'))
            await new_task(consumer(queue, 'cons2'))
            await sleep(0.1)
            for n in range(4):
                await queue.put(n)
                await sleep(0.1)
            for n in range(2):
                await queue.put(None)
            results.append('producer_join')
            await queue.join()
            results.append('producer_done')

        kernel.add_task(producer())
        kernel.run()

        self.assertEqual(results, [
                'producer_start',
                ('cons1', 0),
                ('cons2', 1),
                ('cons1', 2),
                ('cons2', 3),
                'cons1 done',
                'cons2 done',
                'producer_join',
                'producer_done',
                ])

    def test_queue_unbounded(self):
        kernel = get_kernel()
        results = []
        async def consumer(queue, label):
              while True:
                  item = await queue.get()
                  if item is None:
                      break
                  results.append((label, item))
                  await queue.task_done()
              await queue.task_done()
              results.append(label + ' done')

        async def producer():
            queue = Queue()
            results.append('producer_start')
            await new_task(consumer(queue, 'cons1'))
            await sleep(0.1)
            for n in range(4):
                await queue.put(n)
            await queue.put(None)
            results.append('producer_join')
            await queue.join()
            results.append('producer_done')

        kernel.add_task(producer())
        kernel.run()

        self.assertEqual(results, [
                'producer_start',
                ('cons1', 0),
                ('cons1', 1),
                ('cons1', 2),
                ('cons1', 3),
                'cons1 done',
                'producer_join',
                'producer_done',
                ])


    def test_queue_bounded(self):
        kernel = get_kernel()
        results = []
        async def consumer(queue, label):
              while True:
                  item = await queue.get()
                  if item is None:
                      break
                  results.append((label, item))
                  await sleep(0.1)
                  await queue.task_done()
              await queue.task_done()
              results.append(label + ' done')

        async def producer():
            queue = Queue(maxsize=2)
            results.append('producer_start')
            await new_task(consumer(queue, 'cons1'))
            await sleep(0.1)
            for n in range(4):
                await queue.put(n)
                results.append(('produced', n))
            await queue.put(None)
            results.append('producer_join')
            await queue.join()
            results.append('producer_done')

        kernel.add_task(producer())
        kernel.run()

        self.assertEqual(results, [
                'producer_start',
                ('cons1', 0),
                ('produced', 0),
                ('produced', 1),
                ('produced', 2),
                ('produced', 3),
                ('cons1', 1),
                'producer_join',
                ('cons1', 2),
                ('cons1', 3),
                'producer_done',
                'cons1 done',
                ])

    def test_queue_get_cancel(self):
        # Make sure a blocking get can be cancelled
        kernel = get_kernel()
        results = []
        async def consumer():
              queue = Queue()
              try:
                  results.append('consumer waiting')
                  item = await queue.get()
                  results.append('not here')
              except CancelledError:
                  results.append('consumer cancelled')

        async def driver():
            task = await new_task(consumer())
            await sleep(0.5)
            await task.cancel()

        kernel.add_task(driver())
        kernel.run()
        self.assertEqual(results, [
                'consumer waiting',
                'consumer cancelled'
                ])

    def test_queue_put_cancel(self):
        # Make sure a blocking put() can be cancelled
        kernel = get_kernel()
        results = []

        async def producer():
            queue = Queue(1)
            results.append('producer_start')
            await queue.put(0)
            try:
                await queue.put(1)
                results.append('not here')
            except CancelledError:
                results.append('producer_cancel')

        async def driver():
            task = await new_task(producer())
            await sleep(0.5)
            await task.cancel()

        kernel.add_task(driver())
        kernel.run()
        self.assertEqual(results, [
                'producer_start',
                'producer_cancel'
                ])


    def test_queue_get_timeout(self):
        # Make sure a blocking get respects timeouts
        kernel = get_kernel()
        results = []
        async def consumer():
              queue = Queue()
              try:
                  results.append('consumer waiting')
                  item = await queue.get(timeout=0.5)
                  results.append('not here')
              except TimeoutError:
                  results.append('consumer timeout')

        kernel.add_task(consumer())
        kernel.run()
        self.assertEqual(results, [
                'consumer waiting',
                'consumer timeout'
                ])

    def test_queue_put_timeout(self):
        # Make sure a blocking put() respects timeouts
        kernel = get_kernel()
        results = []

        async def producer():
            queue = Queue(1)
            results.append('producer start')
            await queue.put(0)
            try:
                await queue.put(1, timeout=0.5)
                results.append('not here')
            except TimeoutError:
                results.append('producer timeout')

        kernel.add_task(producer())
        kernel.run()
        self.assertEqual(results, [
                'producer start',
                'producer timeout'
                ])

    '''
    def test_queue_put_nowait(self):
        q = Queue(1)
        q.put_nowait(1)
        self.assertTrue(q.full())
        self.assertEqual(q.qsize(), 1)
        with self.assertRaises(QueueFull):
            q.put_nowait(2)

    def test_queue_get_nowait(self):
        q = Queue()
        self.assertTrue(q.empty())
        with self.assertRaises(QueueEmpty):
            q.get_nowait()
    '''

if __name__ == '__main__':
    unittest.main()



# curio/monitor.py
#
# Copyright (C) 2015
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# Debugging monitor for curio.   At the moment, this is overly simplistic.
# It could be expanded to do a lot of other stuff later.

import sys
import os
import traceback
import linecache
import signal
import time
import atexit

from .io import Stream
from .kernel import _kernel_reference, SignalSet, CancelledError
from . import socket

def get_stack(task):
    frames = []
    coro = task.coro
    while coro:
        f = coro.cr_frame if hasattr(coro, 'cr_frame') else coro.gi_frame
        if f is not None:
            frames.append(f)
        coro = coro.cr_await if hasattr(coro, 'cr_await') else coro.gi_yieldfrom
    return frames

def print_stack(task):
        extracted_list = []
        checked = set()
        for f in get_stack(task):
            lineno = f.f_lineno
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            if filename not in checked:
                checked.add(filename)
                linecache.checkcache(filename)
            line = linecache.getline(filename, lineno, f.f_globals)
            extracted_list.append((filename, lineno, name, line))
        if not extracted_list:
            print('No stack for %r' % task)
        else:
            print('Stack for %r (most recent call last):' % task)
        traceback.print_list(extracted_list)

    # Debugging
def ps(kernel):
    headers = ('Task', 'State', 'Cycles', 'Timeout', 'Task')
    widths = (6, 12, 10, 7, 50)
    for h, w in zip(headers, widths):
        print('%-*s' % (w, h), end=' ')
    print()
    print(' '.join(w*'-' for w in widths))
    timestamp = time.monotonic()
    for taskid in sorted(kernel._tasks):
        task = kernel._tasks[taskid]
        remaining = format((task.timeout - timestamp), '0.6f')[:7] if task.timeout else 'None'
        print('%-*d %-*s %-*d %-*s %-*s' % (widths[0], taskid, 
                                            widths[1], task.state,
                                            widths[2], task.cycles,
                                            widths[3], remaining,
                                            widths[4], task))

async def monrun(kernel):
    stdin = Stream(sys.stdin.buffer.raw)
    try:
         print('\nCurio Monitor:  %d tasks running' % len(kernel._tasks))
         print('Type help for commands')

         while True:
              print('curio > ', end='', flush=True)
              resp = await stdin.readline()
              if not resp or resp.startswith(b'q'):
                   print('Leaving monitor')
                   return
              elif resp.startswith(b'p'):
                   ps(kernel)
              elif resp.startswith(b'exit'):
                   raise SystemExit()
              elif resp.startswith(b'cancel'):
                   try:
                        _, taskid_s = resp.split()
                        taskid = int(taskid_s)
                        if taskid in kernel._tasks:
                             print('Cancelling task', taskid)
                             await kernel._tasks[taskid].cancel()
                        else:
                             print('Bad task id')
                   except Exception as e:
                        print('Bad command')
              elif resp.startswith(b'signal'):
                  try:
                      _, signame = resp.split()
                      signame = signame.decode('ascii').strip()
                      if hasattr(signal, signame):
                          os.kill(os.getpid(), getattr(signal, signame))
                  except Exception as e:
                      print('Bad command',e )

              elif resp.startswith(b'w'):
                   try:
                        _, taskid_s = resp.split()
                        taskid = int(taskid_s)
                        if taskid in kernel._tasks:
                             print_stack(kernel._tasks[taskid])
                             print()
                   except Exception as e:
                        print('Bad command', e)
                             
              elif resp.startswith(b'h'):
                   print(
     '''Commands:
         ps               : Show task table
         where taskid     : Show stack frames for a task
         cancel taskid    : Cancel a task
         signal signame   : Send a Unix signal
         exit             : Raise SystemExit and terminate
         quit             : Leave the monitor
     ''')
              elif resp == b'\n':
                   pass
              else:
                   print('Unknown command. Type help.')

    finally:
         os.set_blocking(stdin.fileno(), True)


async def monitor():
    kernel = await _kernel_reference()
    # If the kernel monitor is being used, it's critical that it be shutdown cleanly at exit
    atexit.register(kernel.shutdown)

    sigset = SignalSet(signal.SIGINT)
    while True:
        await sigset.wait()
        with sigset.ignore():
            await monrun(kernel)


async def webmonitor(address):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
    sock.listen(1)
    while True:
        client, addr = await sock.accept()
        await new_task(webmonitor_handler(client))

async def webmonitor_handler(client):
    request = []
    async with client.makefile('rb') as client_f:
        async for line in client_f:
            if not line.strip():
                break
            request.append(line)

    # Look at the request line
    meth, path, proto = request[0].split()
    
             
    



# kernel.py
# 
# Copyright (C) 2015
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# This is the core of curio.   Definitions for tasks, signal sets, and the kernel
# are here.

import socket
import heapq
import time
import os
import sys
import logging
import inspect
import signal

from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE
from collections import deque, defaultdict
from types import coroutine
from contextlib import contextmanager

# Logger where uncaught exceptions from crashed tasks are logged
log = logging.getLogger(__name__)

# kqueue is the datatype used by the kernel for all of its queuing functionality.
# Any time a task queue is needed, use this type instead of directly hard-coding the
# use of a deque.  This will make sure the code continues to work even if the
# queue type is changed later.

kqueue = deque

# --- Curio specific exceptions

class CurioError(Exception):
    pass

class CancelledError(CurioError):
    pass

class TaskError(CurioError):
    pass

class _CancelRetry(Exception):
    pass

# Task class wraps a coroutine, but provides other information about 
# the task itself for the purposes of debugging, scheduling, timeouts, etc.

class Task(object):
    __slots__ = ('id', 'parent_id', 'children', 'coro', 'cycles', 'state',
                 'cancel_func', 'future', 'timeout', 'exc_info', 'next_value',
                 'next_exc', 'joining', 'terminated', '__weakref__')
    _lastid = 1
    def __init__(self, coro):
        self.id = Task._lastid
        Task._lastid += 1
        self.parent_id = None     # Parent task id
        self.children = set()     # Set of child tasks
        self.coro = coro          # Underlying generator/coroutine
        self.cycles = 0           # Execution cycles completed
        self.state = 'INITIAL'    # Execution state
        self.cancel_func = None   # Cancellation function
        self.future = None        # Pending Future (if any)
        self.timeout = None       # Pending timeout (if any)
        self.exc_info = None      # Exception info (if any on crash)
        self.next_value = None    # Next value to send on execution
        self.next_exc = None      # Next exception to send on execution
        self.joining = None       # Optional set of tasks waiting to join with this one
        self.terminated = False   # Terminated?

    def __repr__(self):
        return 'Task(id=%r, %r, state=%r)' % (self.id, self.coro, self.state)

    def __str__(self):
        return self.coro.__qualname__

    async def join(self, *, timeout=None):
        '''
        Waits for a task to terminate.  Returns the return value (if any)
        or raises a TaskError if the task crashed with an exception.
        '''
        await _join_task(self, timeout)
        if self.exc_info:
            raise TaskError('Task crash') from self.exc_info[1]
        else:
            return self.next_value

    async def cancel(self, *, exc=CancelledError, timeout=None):
        '''
        Cancels a task.  Does not return until the task actually terminates.
        '''
        await _cancel_task(self, exc, timeout)

    async def cancel_children(self, *, exc=CancelledError, timeout=None):
        '''
        Cancels all of the children of this task.
        '''
        for task in list(self.children):
            await _cancel_task(task, exc, timeout=timeout)

# The SignalSet class represents a set of Unix signals being monitored. 
class SignalSet(object):
    def __init__(self, *signos):
        self.signos = signos             # List of all signal numbers being tracked
        self.pending = deque()           # Pending signals received
        self.waiting = None              # Task waiting for the signals (if any)
        self.watching = False            # Are the signals being watched right now?

    
    async def __aenter__(self):
        assert not self.watching
        await _sigwatch(self)
        self.watching = True
        return self

    async def __aexit__(self, *args):
        await _sigunwatch(self)
        self.watching = False

    async def wait(self, *, timeout=None):
        '''
        Wait for a single signal from the signal set to arrive.
        '''
        if not self.watching:
            async with self:
                return await self.wait(timeout=timeout)

        while True:
            if self.pending:
                return self.pending.popleft()
            await _sigwait(self, timeout)

    @contextmanager
    def ignore(self):
        '''
        Context manager. Temporarily ignores all signals in the signal set. 
        '''
        try:
            orig_signals = [ (signo, signal.signal(signo, signal.SIG_IGN)) for signo in self.signos ]
            yield
        finally:
            for signo, handler in orig_signals:
                signal.signal(signo, handler)

# Underlying kernel that drives everything

class Kernel(object):
    __slots__ = ('_selector', '_ready', '_tasks', '_current', '_sleeping',
                 '_signals', '_default_signals', '_njobs', '_running',
                 '_notify_sock', '_wait_sock')

    def __init__(self, selector=None, with_monitor=False):
        if selector is None:
            selector = DefaultSelector()
        self._selector = selector
        self._ready = kqueue()      # Tasks ready to run
        self._tasks = { }           # Task table
        self._current = None        # Current task
        self._sleeping = []         # Heap of tasks waiting on a timer for any reason
        self._signals = None        # Dict { signo: [ sigsets ] } of watched signals (initialized only if signals used)
        self._default_signals = {}  # Dict of default signal handlers
        self._njobs = 0             # Number of non-daemonic jobs running
        self._running = False       # Kernel running?

        # Create the init task responsible for waking the event loop and processing signals
        self._notify_sock, self._wait_sock = socket.socketpair()
        self._wait_sock.setblocking(False)
        self._notify_sock.setblocking(False)
        self.add_task(self._init_task(), daemon=True)

        if with_monitor and sys.stdin.isatty():
            from .monitor import monitor
            self.add_task(monitor(), daemon=True)

    def __del__(self):
        self._notify_sock.close()
        self._wait_sock.close()

    # Force the kernel to wake.  Used on non-I/O events such as completion of futures
    def _wake(self, task=None, value=None, exc=None):
        if task:
            self._reschedule_task(task, value, exc)
        self._notify_sock.send(b'\x00')

    # Internal task that monitors the loopback socket--allowing the kernel to
    # awake for non-I/O events.  Also processes incoming signals.
    async def _init_task(self):
        while True:
            await _read_wait(self._wait_sock)
            data = self._wait_sock.recv(100)

            # Any non-null bytes received here are assumed to be received signals.
            # See if there are any pending signal sets and unblock if needed 
            if not self._signals:
                continue

            signals = (signal.Signals(n) for n in data if n in self._signals)
            for signo in signals:
                for sigset in self._signals[signo]:
                    sigset.pending.append(signo)
                    if sigset.waiting:
                        self._reschedule_task(sigset.waiting, value=signo)
                        sigset.waiting = None

    # Public task management functions.  These are the only methods safe to use
    # outside of the kernel loop. They must be called when the loop is stopped.

    def add_task(self, coro, daemon=False):
        assert self._current is None
        return self._new_task(coro, daemon)

    # Internal task management functions. 
    def _new_task(self, coro, daemon=False):
        # Create a new task in the kernel.  If daemon is True, the task is
        # created with no parent task.

        task = Task(coro)
        self._tasks[task.id] = task
        self._reschedule_task(task)
        if not daemon:
            self._njobs += 1
            ptask = self._current if self._current else self._tasks[1]
            task.parent_id = ptask.id
            ptask.children.add(task)
        return task

    def _reschedule_task(self, task, value=None, exc=None):
        # Reschedule a task, putting it back on the ready queue so that it can run.
        # value and exc specify a value or exception to send into the underlying 
        # coroutine when it is rescheduled.

        assert task.id in self._tasks, 'Task %r not in the task table' % task
        self._ready.append(task)
        task.next_value = value
        task.next_exc = exc
        task.state = 'READY'
        task.timeout = task.cancel_func = task.future = None

    def _cancel_task(self, task, exc=CancelledError):
        # Cancel a task. This causes a CancelledError exception to raise in the
        # underlying coroutine.  The coroutine can elect to catch the exception
        # and continue to run to perform cleanup actions.  However, it would 
        # normally terminate shortly afterwards.   This method is also used to 
        # raise timeouts.

        assert task != self._current, "A task can't cancel itself (%r, %r)" % (task, self._current)
        
        if task.terminated:
            return True

        if not task.cancel_func:
            # If there is no cancellation function set. It means that
            # the task finished whatever it might have been working on
            # and it's sitting in the ready queue ready to run.  This
            # presents a rather tricky corner case of
            # cancellation. First of all, it means that the task
            # successfully completed whatever it was waiting to do,
            # but it has not communicated the result back.  This could
            # be something critical like acquiring a lock.  Because of
            # that, can't just arbitrarily nuke the task.  Instead, we
            # have to let it be rescheduled and properly process the result of the
            # successfully completed operation.   The return of False
            # here means that cancellation failed and that it should be retried.
            return False

        # Detach the task from where it might be waiting at this moment
        task.cancel_func()
        
        assert task not in self._ready
        # Reschedule it with a pending exception
        self._reschedule_task(task, exc=exc())
        return True

    def _cleanup_task(self, task, value=None, exc=None):
        # Cleanup task.  This is called after the underlying coroutine has
        # terminated.  value and exc give the return value or exception of
        # the coroutine.  This wakes any tasks waiting to join and removes
        # the terminated task from its parent/children.

        task.next_value = value
        if task.parent_id:
            self._njobs -=1

        if task.joining:
            for wtask in task.joining:
                self._reschedule_task(wtask, value=value, exc=exc)
            task.joining = None
        task.terminated = True
        del self._tasks[task.id]

        # Remove the task from the parent child list
        parent = self._tasks.get(task.parent_id)
        if parent:
            parent.children.remove(task)

        # Reassign all remaining children to the parent task
        init = self._tasks[1]
        init.children.update(task.children)
        for ctask in task.children:
            ctask.parent_id = 1
        task.children = set()

    def _set_timeout(self, task, seconds, sleep_type='timeout'):
        task.timeout = time.monotonic() + seconds
        item = (task.timeout, task, sleep_type)
        heapq.heappush(self._sleeping, item)

    # I/O 
    def _poll_for_io(self):
        if self._sleeping:
            timeout = self._sleeping[0][0] - time.monotonic()
        else:
            timeout = None

        events = self._selector.select(timeout)
        for key, mask in events:
            task = key.data
            self._selector.unregister(key.fileobj)
            self._reschedule_task(task)

        # Process sleeping tasks (if any)
        if self._sleeping:
            current = time.monotonic()
            while self._sleeping and self._sleeping[0][0] <= current:
                tm, task, sleep_type = heapq.heappop(self._sleeping)
                # When a task wakes, verify that the timeout value matches that stored
                # on the task. If it differs, it means that the task completed its
                # operation, was cancelled, or is no longer concerned with this
                # sleep operation.  In that case, we do nothing
                if tm == task.timeout:
                    if sleep_type == 'sleep':
                        self._reschedule_task(task)
                    elif sleep_type == 'timeout':
                        self._cancel_task(task, exc=TimeoutError)

    # Kernel central loop
    def run(self, coro=None, *, pdb=False, log_errors=True):
        '''
        Run the kernel until no more non-daemonic tasks remain.
        '''

        # If a coroutine was given, add it as the first task
        if coro:
            self.add_task(coro)

        self._running = True

        while self._njobs > 0 and self._running:
            self._current = None

            # Poll for I/O as long as there is nothing to run
            while not self._ready:
                self._poll_for_io()

            # Run everything that's ready
            while self._ready:
                current = self._current = self._ready.popleft()
                assert current.id in self._tasks
                try:
                    current.state = 'RUNNING'
                    current.cycles += 1
                    if current.next_exc is None:
                        op, *args = current.coro.send(current.next_value)
                    else:
                        op, *args = current.coro.throw(current.next_exc)

                    # Execute the trap
                    getattr(self, op)(current, *args)

                except (StopIteration, CancelledError) as e:
                    self._cleanup_task(current, value = e.value if isinstance(e, StopIteration) else None)

                except Exception as e:
                    current.exc_info = sys.exc_info()
                    current.state = 'CRASHED'
                    exc = TaskError('Task Crashed')
                    exc.__cause__ = e
                    self._cleanup_task(current, exc=exc)
                    if log_errors:
                        log.error('Curio: Task Crash: %s' % current, exc_info=True)
                    if pdb:
                        import pdb as _pdb
                        _pdb.post_mortem(current.exc_info[2])

                except SystemExit:
                    self._cleanup_task(current)
                    raise


        self._current = None

    def stop(self):
        '''
        Stop the kernel loop.
        '''
        self._running = False

    def shutdown(self):
        '''
        Cleanly shut down the kernel.  All remaining tasks are cancelled.  Using the
        function is highly recommended prior to terminating any program.
        '''
        self._current = None
        for task in sorted(self._tasks.values(), key=lambda t: t.id, reverse=True):
            if task.id == 1:
                continue
            # If the task is daemonic, force it to non-daemon status and cancel it
            if not task.parent_id:
                self._njobs += 1
                task.parent_id = -1

            assert self._cancel_task(task)

        self.run()

    # Traps.  These implement the low-level functionality that is triggered by coroutines.
    # You shouldn't invoke these directly. Instead, coroutines use a statement such as
    #   
    #   yield ('_trap_io', sock, EVENT_READ, 'READ_WAIT', timeout)
    #
    # To execute these methods.

    def _trap_io(self, current, fileobj, event, state, timeout):
        self._selector.register(fileobj, event, current)
        current.state = state
        current.cancel_func = lambda: self._selector.unregister(fileobj)
        if timeout:
            self._set_timeout(current, timeout)

    def _trap_future_wait(self, current, future, timeout=None):
        current.state = 'FUTURE_WAIT'
        current.cancel_func = lambda: future.cancel()
        current.future = future
        future.add_done_callback(lambda fut, task=current: self._wake(task) if task.future == fut else None)
        if timeout:
            self._set_timeout(current, timeout)

    def _trap_new_task(self, current, coro, daemon=False):
        task = self._new_task(coro, daemon)
        self._reschedule_task(current, value=task)

    def _trap_reschedule_tasks(self, current, queue, n=1, value=None, exc=None):
        while n > 0:
            self._reschedule_task(queue.popleft(), value=value, exc=exc)
            n -= 1
        self._reschedule_task(current)

    def _trap_join_task(self, current, task, timeout=None):
        if task.terminated:
            self._reschedule_task(current)
        else:
            if task.joining is None:
                task.joining = kqueue()
            self._trap_wait_queue(current, task.joining, 'TASK_JOIN', timeout)

    def _trap_cancel_task(self, current, task, exc, timeout=None):
        if self._cancel_task(task, exc):
            self._trap_join_task(current, task, timeout)
        else:
            # Fail with a _CancelRetry exception to indicate that the cancel
            # request should be attempted again.  This happens in the case
            # that a cancellation request is issued against a task that
            # ready to run in the ready queue.
            self._reschedule_task(current, exc=_CancelRetry())

    def _trap_wait_queue(self, current, queue, state, timeout=None):
        queue.append(current)
        current.state = state
        current.cancel_func = lambda: queue.remove(current)
        if timeout:
            self._set_timeout(current, timeout)
        
    def _trap_sleep(self, current, seconds):
        if seconds > 0:
            self._set_timeout(current, seconds, 'sleep')
            current.state = 'TIME_SLEEP'
            current.cancel_func = lambda: None
        else:
            self._reschedule_task(current)

    def _trap_sigwatch(self, current, sigset):
        # Initialize the signal handling part of the kernel if not done already
        # Note: This only works if running in the main thread
        if self._signals is None:
            self._signals = defaultdict(list)
            signal.set_wakeup_fd(self._notify_sock.fileno())     
            
        for signo in sigset.signos:
            if not self._signals[signo]:
                self._default_signals[signo] = signal.signal(signo, lambda signo, frame:None)
            self._signals[signo].append(sigset)

        self._reschedule_task(current)
        
    def _trap_sigunwatch(self, current, sigset):
        for signo in sigset.signos:
            if sigset in self._signals[signo]:
                self._signals[signo].remove(sigset)

            # If there are no active watchers for a signal, revert it back to default behavior
            if not self._signals[signo]:
                signal.signal(signo, self._default_signals[signo])
                del self._signals[signo]
        self._reschedule_task(current)

    def _trap_sigwait(self, current, sigset, timeout):
        sigset.waiting = current
        current.state = 'SIGNAL_WAIT'
        current.cancel_func = lambda: setattr(sigset, 'waiting', None)
        if timeout:
            self._set_timeout(current, timeout)

    def _trap_kernel(self, current):
        self._reschedule_task(current, value=self)

# Coroutines corresponding to the kernel traps.  These functions provide the bridge from
# tasks to the underlying kernel. This is the only place with the explicit @coroutine
# decorator needs to be used.  All other code in curio and in user-tasks should use
# async/await instead.   Direct use by users is allowed, but if you're working with
# these traps directly, there is probably a higher level interface that simplifies
# the problem you're trying to solve (e.g., Socket, File, objects, etc.)

@coroutine
def _read_wait(fileobj, timeout=None):
    '''
    Wait until reading can be performed.
    '''
    yield ('_trap_io', fileobj, EVENT_READ, 'READ_WAIT', timeout)

@coroutine
def _write_wait(fileobj, timeout=None):
    '''
    Wait until writing can be performed.
    '''
    yield ('_trap_io', fileobj, EVENT_WRITE, 'WRITE_WAIT', timeout)

@coroutine
def _future_wait(future, timeout=None):
    '''
    Wait for the future of a Future to be computed.
    '''
    yield ('_trap_future_wait', future, timeout)

@coroutine
def _sleep(seconds):
    '''
    Sleep for a given number of seconds
    '''
    yield ('_trap_sleep', seconds)

@coroutine
def _new_task(coro, daemon):
    '''
    Create a new task
    '''
    return (yield '_trap_new_task', coro, daemon)

@coroutine
def _cancel_task(task, exc=CancelledError, timeout=None):
    '''
    Cancel a task.  Causes a CancelledError exception to raise in the task.
    '''
    while True:
        try:
            yield ('_trap_cancel_task', task, exc, timeout)
            return
        except _CancelRetry:
            pass

@coroutine
def _join_task(task, timeout=None):
    '''
    Wait for a task to terminate.
    '''
    yield ('_trap_join_task', task, timeout)

@coroutine
def _wait_on_queue(queue, state, timeout=None):
    '''
    Put the task to sleep on a kernel queue.
    '''
    yield ('_trap_wait_queue', queue, state, timeout)

@coroutine
def _reschedule_tasks(queue, n=1, value=None, exc=None):
    '''
    Reschedule one or more tasks waiting on a kernel queue.
    '''
    yield ('_trap_reschedule_tasks', queue, n, value, exc)

@coroutine
def _sigwatch(sigset):
    '''
    Start monitoring a signal set
    '''
    yield ('_trap_sigwatch', sigset)

@coroutine
def _sigunwatch(sigset):
    '''
    Stop watching a signal set
    '''
    yield ('_trap_sigunwatch', sigset)

@coroutine
def _sigwait(sigset, timeout=None):
    '''
    Wait for a signal to arrive.
    '''
    yield ('_trap_sigwait', sigset, timeout)

@coroutine
def _kernel_reference():
    '''
    Get a reference to the running kernel
    '''
    result = yield ('_trap_kernel',)
    return result

# Public-facing syscalls.  These coroutines wrap a low-level coroutine
# with an extra layer involving an async/await function.  The main
# reason for doing this is that the user will get proper warning
# messages if they forget to use the required 'await' keyword.

async def sleep(seconds):
    '''
    Sleep for a specified number of seconds.
    '''
    await _sleep(seconds)

async def new_task(coro, *, daemon=False):
    '''
    Create a new task.
    '''
    return await _new_task(coro, daemon)

_default_kernel = None
def get_kernel():
    '''
    Return the default kernel.
    '''
    global _default_kernel
    if _default_kernel is None:
        _default_kernel = Kernel()
    return _default_kernel

__all__ = [ 'Kernel', 'get_kernel', 'sleep', 'new_task', 'SignalSet', 'TaskError', 'CancelledError' ]
            
        



from .sync import *
from .queue import *
from .kernel import *
from .socket import *
from .subprocess import *
from .workers import *
from .network import *

if __name__ == '__main__':
    import unittest
    unittest.main()
    



# sync.py
#
# Different test scenarios designed to run under management of a kernel

import unittest
from collections import deque

from ..import *

# ---- Synchronization primitives

class TestEvent(unittest.TestCase):
    def test_event_get_wait(self):
        kernel = get_kernel()
        results = []
        async def event_setter(evt, seconds):
              results.append('sleep')
              await sleep(seconds)
              results.append('event_set')
              await evt.set()

        async def event_waiter(evt):
              results.append('wait_start')
              results.append(evt.is_set())
              await evt.wait()
              results.append('wait_done')
              results.append(evt.is_set())
              evt.clear()
              results.append(evt.is_set())

        evt = Event()
        kernel.add_task(event_waiter(evt))
        kernel.add_task(event_setter(evt, 1))
        kernel.run()
        self.assertEqual(results, [
                'wait_start',
                False,
                'sleep',
                'event_set',
                'wait_done',
                True,
                False
                ])

    def test_event_get_immediate(self):
        kernel = get_kernel()
        results = []
        async def event_setter(evt):
              results.append('event_set')
              await evt.set()

        async def event_waiter(evt, seconds):
              results.append('sleep')
              await sleep(seconds)
              results.append('wait_start')
              await evt.wait()
              results.append('wait_done')

        evt = Event()
        kernel.add_task(event_waiter(evt, 1))
        kernel.add_task(event_setter(evt))
        kernel.run()
        self.assertEqual(results, [
                'sleep',
                'event_set',
                'wait_start',
                'wait_done',
                ])


    def test_event_wait_cancel(self):
        kernel = get_kernel()
        results = []
        async def event_waiter(evt):
              results.append('event_wait')
              try:
                   await evt.wait()
              except CancelledError:
                   results.append('event_cancel')

        async def event_cancel(seconds):
              evt = Event()
              task = await new_task(event_waiter(evt))
              results.append('sleep')
              await sleep(seconds)
              results.append('cancel_start')
              await task.cancel()
              results.append('cancel_done')

        kernel.add_task(event_cancel(1))
        kernel.run()
        self.assertEqual(results, [
                'event_wait',
                'sleep',
                'cancel_start',
                'event_cancel',
                'cancel_done',
                ])

    def test_event_wait_timeout(self):
        kernel = get_kernel()
        results = []
        async def event_waiter(evt):
              results.append('event_wait')
              try:
                   await evt.wait(timeout=0.5)
              except TimeoutError:
                   results.append('event_timeout')

        async def event_run(seconds):
              evt = Event()
              task = await new_task(event_waiter(evt))
              results.append('sleep')
              await sleep(seconds)
              results.append('sleep_done')

        kernel.add_task(event_run(1))
        kernel.run()
        self.assertEqual(results, [
                'event_wait',
                'sleep',
                'event_timeout',
                'sleep_done',
                ])


    def test_event_wait_notimeout(self):
        kernel = get_kernel()
        results = []
        async def event_waiter(evt):
              results.append('event_wait')
              try:
                   await evt.wait(timeout=1.0)
                   results.append('got event')
              except TimeoutError:
                   results.append('event_timeout')
              
              evt.clear()
              try:
                  await evt.wait()
                  results.append('got event')
              except TimeoutError:
                  results.append('bad timeout')

        async def event_run():
              evt = Event()
              task = await new_task(event_waiter(evt))
              results.append('sleep')
              await sleep(0.25)
              results.append('event_set')
              await evt.set()
              await sleep(1.0)
              results.append('event_set')
              await evt.set()

        kernel.add_task(event_run())
        kernel.run()
        self.assertEqual(results, [
                'event_wait',
                'sleep',
                'event_set',
                'got event',
                'event_set',
                'got event'
                ])

class TestLock(unittest.TestCase):
    def test_lock_sequence(self):
        kernel = get_kernel()
        results = []
        async def worker(lck, label):
              results.append(label + ' wait')
              results.append(lck.locked())
              async with lck:
                   results.append(label + ' acquire')
                   await sleep(0.25)
              results.append(label + ' release')

        lck = Lock()
        kernel.add_task(worker(lck, 'work1'))
        kernel.add_task(worker(lck, 'work2'))
        kernel.add_task(worker(lck, 'work3'))
        kernel.run()
        self.assertEqual(results, [
                'work1 wait',
                False,
                'work1 acquire',
                'work2 wait',
                True,
                'work3 wait',
                True,
                'work2 acquire',
                'work1 release',
                'work3 acquire',
                'work2 release',
                'work3 release',
                ])

    def test_lock_acquire_cancel(self):
        kernel = get_kernel()
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  async with lck:
                       results.append('never here')
              except CancelledError:
                  results.append('lock_cancel')

        async def worker_cancel(seconds):
              lck = Lock()
              async with lck:
                  task = await new_task(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('cancel_start')
                  await task.cancel()
                  results.append('cancel_done')

        kernel.add_task(worker_cancel(1))
        kernel.run()
        self.assertEqual(results, [
                'lock_wait',
                'sleep',
                'cancel_start',
                'lock_cancel',
                'cancel_done',
                ])


    def test_lock_acquire_timeout(self):
        kernel = get_kernel()
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  await lck.acquire(timeout=0.5)
                  results.append('never here')
                  await lck.release()
              except TimeoutError:
                  results.append('lock_timeout')

        async def worker_timeout(seconds):
              lck = Lock()
              async with lck:
                  await new_task(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('sleep_done')

        kernel.add_task(worker_timeout(1))
        kernel.run()
        self.assertEqual(results, [
                'lock_wait',
                'sleep',
                'lock_timeout',
                'sleep_done',
                ])


class TestSemaphore(unittest.TestCase):
    def test_sema_sequence(self):
        kernel = get_kernel()
        results = []
        async def worker(sema, label):
              results.append(label + ' wait')
              results.append(sema.locked())
              async with sema:
                   results.append(label + ' acquire')
                   await sleep(0.25)
              results.append(label + ' release')

        sema = Semaphore()
        kernel.add_task(worker(sema, 'work1'))
        kernel.add_task(worker(sema, 'work2'))
        kernel.add_task(worker(sema, 'work3'))
        kernel.run()
        self.assertEqual(results, [
                'work1 wait',
                False,
                'work1 acquire',
                'work2 wait',
                True,
                'work3 wait',
                True,
                'work2 acquire',
                'work1 release',
                'work3 acquire',
                'work2 release',
                'work3 release',
                ])

    def test_sema_sequence2(self):
        kernel = get_kernel()
        results = []
        async def worker(sema, label, seconds):
              results.append(label + ' wait')
              results.append(sema.locked())
              async with sema:
                   results.append(label + ' acquire')
                   await sleep(seconds)
              results.append(label + ' release')

        sema = Semaphore(2)
        kernel.add_task(worker(sema, 'work1', 0.25))
        kernel.add_task(worker(sema, 'work2', 0.30))
        kernel.add_task(worker(sema, 'work3', 0.35))
        kernel.run()
        self.assertEqual(results, [
                'work1 wait',            # Both work1 and work2 admitted
                False,
                'work1 acquire',
                'work2 wait',
                False,
                'work2 acquire',
                'work3 wait',
                True,
                'work3 acquire',
                'work1 release',
                'work2 release',
                'work3 release',
                ])

    def test_sema_acquire_cancel(self):
        kernel = get_kernel()
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  async with lck:
                       results.append('never here')
              except CancelledError:
                  results.append('lock_cancel')

        async def worker_cancel(seconds):
              lck = Semaphore()
              async with lck:
                  task = await new_task(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('cancel_start')
                  await task.cancel()
                  results.append('cancel_done')

        kernel.add_task(worker_cancel(1))
        kernel.run()
        self.assertEqual(results, [
                'lock_wait',
                'sleep',
                'cancel_start',
                'lock_cancel',
                'cancel_done',
                ])

    def test_sema_acquire_timeout(self):
        kernel = get_kernel()
        results = []
        async def worker(lck):
              results.append('lock_wait')
              try:
                  await lck.acquire(timeout=0.5)
                  results.append('never here')
                  await lck.release()
              except TimeoutError:
                  results.append('lock_timeout')

        async def worker_timeout(seconds):
              lck = Semaphore()
              async with lck:
                  await new_task(worker(lck))
                  results.append('sleep')
                  await sleep(seconds)
                  results.append('sleep_done')

        kernel.add_task(worker_timeout(1))
        kernel.run()
        self.assertEqual(results, [
                'lock_wait',
                'sleep',
                'lock_timeout',
                'sleep_done',
                ])

    def test_bounded(self):
        kernel = get_kernel()
        results = []
        async def task():
            sema = BoundedSemaphore(1)
            try:
                await sema.release()
                results.append('not here')
            except ValueError:
                results.append('value error')

        kernel.add_task(task())
        kernel.run()
        self.assertEqual(results, [
                'value error',
                ])


class TestCondition(unittest.TestCase):
    def test_cond_sequence(self):
        kernel = get_kernel()
        results = []
        async def consumer(cond, q, label):
              while True:
                  async with cond:
                      if not q:
                          results.append(label + ' wait')
                          await cond.wait()
                      item = q.popleft()
                      if item is None:
                          break
                      results.append((label, item))
              results.append(label+' done') 

        async def producer(cond, q, count, nproducers):
             for n in range(count):
                 async with cond:
                     q.append(n)
                     results.append(('producing', n))
                     await cond.notify()
                 await sleep(0.1)

             for n in range(nproducers):
                 async with cond:
                     q.append(None)
                     results.append(('ending', n))
                     await cond.notify()
                 await sleep(0.1)
              
        cond = Condition()
        q = deque()
        kernel.add_task(consumer(cond, q, 'cons1'))
        kernel.add_task(consumer(cond, q, 'cons2'))
        kernel.add_task(producer(cond, q, 4, 2))
        kernel.run()
        self.assertEqual(results, [
                'cons1 wait',
                'cons2 wait',
                ('producing', 0),
                ('cons1', 0),
                'cons1 wait',
                ('producing', 1),
                ('cons2', 1),
                'cons2 wait',
                ('producing', 2),
                ('cons1', 2),
                'cons1 wait',
                ('producing', 3),
                ('cons2', 3),
                'cons2 wait',
                ('ending', 0),
                ('cons1 done'),
                ('ending', 1),
                ('cons2 done')
                ])

    def test_cond_wait_cancel(self):
        kernel = get_kernel()
        results = []
        async def worker(cond):
              try:
                  async with cond:
                       results.append('cond_wait')
                       await cond.wait()
                       results.append('never here')
              except CancelledError:
                  results.append('worker_cancel')

        async def worker_cancel(seconds):
              cond = Condition()
              task = await new_task(worker(cond))
              results.append('sleep')
              await sleep(seconds)
              results.append('cancel_start')
              await task.cancel()
              results.append('cancel_done')

        kernel.add_task(worker_cancel(1))
        kernel.run()
        self.assertEqual(results, [
                'cond_wait',
                'sleep',
                'cancel_start',
                'worker_cancel',
                'cancel_done',
                ])

    def test_cond_wait_timeout(self):
        kernel = get_kernel()
        results = []
        async def worker(cond):
              try:
                  async with cond:
                       results.append('cond_wait')
                       await cond.wait(timeout=0.25)
                       results.append('never here')
              except TimeoutError:
                  results.append('worker_timeout')

        async def worker_cancel(seconds):
              cond = Condition()
              task = await new_task(worker(cond))
              results.append('sleep')
              await sleep(seconds)
              results.append('done')

        kernel.add_task(worker_cancel(1))
        kernel.run()
        self.assertEqual(results, [
                'cond_wait',
                'sleep',
                'worker_timeout',
                'done'
                ])

    def test_cond_notify_all(self):
        kernel = get_kernel()
        results = []
        async def worker(cond):
             async with cond:
                 results.append('cond_wait')
                 await cond.wait()
                 results.append('wait_done')

        async def worker_notify(seconds):
              cond = Condition()
              await new_task(worker(cond))
              await new_task(worker(cond))
              await new_task(worker(cond))
              results.append('sleep')
              await sleep(seconds)
              async with cond:
                  results.append('notify')
                  await cond.notify_all()
              results.append('done')

        kernel.add_task(worker_notify(1))
        kernel.run()
        self.assertEqual(results, [
                'cond_wait',
                'cond_wait',
                'cond_wait',
                'sleep',
                'notify',
                'wait_done',
                'done',
                'wait_done',
                'wait_done',
                ])

    def test_cond_waitfor(self):
        kernel = get_kernel()
        results = []
        async def consumer(cond, q, label):
             async with cond:
                 results.append(label + ' waitfor')
                 await cond.wait_for(lambda: len(q) > 2)
                 results.append((label, len(q)))
             results.append(label+' done') 

        async def producer(cond, q, count):
             for n in range(count):
                 async with cond:
                     q.append(n)
                     results.append(('producing', n))
                     await cond.notify()
                 await sleep(0.1)
              
        cond = Condition()
        q = deque()
        kernel.add_task(consumer(cond, q, 'cons1'))
        kernel.add_task(consumer(cond, q, 'cons2'))
        kernel.add_task(producer(cond, q, 4))
        kernel.run()
        self.assertEqual(results, [
                'cons1 waitfor',
                'cons2 waitfor',
                ('producing', 0),
                ('producing', 1),
                ('producing', 2),
                ('cons1', 3),
                'cons1 done',
                ('producing', 3),
                ('cons2', 4),
                'cons2 done'
                ])

if __name__ == '__main__':
    unittest.main()



# curio/test/workers.py

import unittest
from .. import subprocess
from ..kernel import *

# ---- Test subprocesses and worker task related functions

class TestSubprocess(unittest.TestCase):
    def test_simple(self):
        kernel = get_kernel()
        results = []
        async def subproc():
            out = await subprocess.run(['python3', '-m', 'curio.test.slow'], stdout=subprocess.PIPE)
            results.append(out.stdout)
            results.append(out.returncode)

        kernel.add_task(subproc())
        kernel.run()
        self.assertEqual(results, [
                b't-minus 4\nt-minus 3\nt-minus 2\nt-minus 1\n',
                0,
                ])

    def test_simple_check_output(self):
        kernel = get_kernel()
        results = []
        async def subproc():
            out = await subprocess.check_output(['python3', '-m', 'curio.test.slow'])
            results.append(out)

        kernel.add_task(subproc())
        kernel.run()
        self.assertEqual(results, [
                b't-minus 4\nt-minus 3\nt-minus 2\nt-minus 1\n',
                ])

    def test_bad_cmd(self):
        kernel = get_kernel()
        results = []
        async def subproc():
            try:
                out = await subprocess.run(['python3', '-m', 'curio.test.bad'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
                results.append('what?')
            except subprocess.CalledProcessError:
                results.append('bad command')

        kernel.add_task(subproc())
        kernel.run()
        self.assertEqual(results, [ 'bad command' ])

    def test_bad_cmd_check_output(self):
        kernel = get_kernel()
        results = []
        async def subproc():
            try:
                out = await subprocess.check_output(['python3', '-m', 'curio.test.bad'], stderr=subprocess.STDOUT)
                results.append('what?')
            except subprocess.CalledProcessError:
                results.append('bad command')

        kernel.add_task(subproc())
        kernel.run()
        self.assertEqual(results, [ 'bad command' ])

    def test_timeout(self):
        kernel = get_kernel()
        results = []
        async def subproc():
            try:
                out = await subprocess.run(['python3', '-m', 'curio.test.slow'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=1)
                results.append('what?')
            except subprocess.TimeoutExpired:
                results.append('timeout')

        kernel.add_task(subproc())
        kernel.run()
        self.assertEqual(results, [ 'timeout' ])

if __name__ == '__main__':
    unittest.main()



# curio/test/socket.py

import unittest
from ..import *
from ..socket import *

class TestSocket(unittest.TestCase):
    def test_tcp_echo(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_STREAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.listen(5)
            results.append('accept wait')
            client, addr = await sock.accept()
            results.append('accept done')
            await new_task(handler(client))
            await sock.close()

        async def handler(client):
            results.append('handler start')
            while True:
                results.append('recv wait')
                data = await client.recv(100)
                if not data:
                    break
                results.append(('handler', data))
                await client.sendall(data)
            results.append('handler done')
            await client.close()

        async def client(address):
            results.append('client start')
            sock = socket(AF_INET, SOCK_STREAM)
            await sock.connect(address)
            await sock.send(b'Msg1')
            await sleep(0.1)
            resp = await sock.recv(100)
            results.append(('client', resp))
            await sock.send(b'Msg2')
            await sleep(0.1)
            resp = await sock.recv(100)
            results.append(('client', resp))
            results.append('client close')
            await sock.close()

        kernel.add_task(server(('',25000)))
        kernel.add_task(client(('localhost', 25000)))
        kernel.run()

        self.assertEqual(results, [
                'accept wait',
                'client start',
                'accept done',
                'handler start',
                'recv wait',
                ('handler', b'Msg1'),
                'recv wait',
                ('client', b'Msg1'),
                ('handler', b'Msg2'),
                'recv wait',
                ('client', b'Msg2'),
                'client close',
                'handler done'
                ])

    def test_tcp_file_echo(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_STREAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.listen(5)
            results.append('accept wait')
            client, addr = await sock.accept()
            results.append('accept done')
            await new_task(handler(client))
            await sock.close()

        async def handler(client):
            results.append('handler start')
            async with client.makefile('wb') as out_f:
                async with client.makefile('rb') as in_f:
                    async for line in in_f:
                        results.append(('handler', line))
                        await out_f.write(line)
                        await out_f.flush()
            results.append('handler done')
            await client.close()

        async def client(address):
            results.append('client start')
            sock = socket(AF_INET, SOCK_STREAM)
            await sock.connect(address)
            in_f = sock.makefile('rb')
            out_f = sock.makefile('wb')
            await out_f.write(b'Msg1\n')
            await out_f.flush()
            await sleep(0.1)
            resp = await in_f.read(100)
            results.append(('client', resp))
            await out_f.write(b'Msg2\n')
            await out_f.flush()
            await sleep(0.1)
            resp = await in_f.read(100)
            results.append(('client', resp))
            results.append('client close')
            await out_f.close()
            await in_f.close()
            await sock.close()

        kernel.add_task(server(('',25000)))
        kernel.add_task(client(('localhost', 25000)))
        kernel.run()

        self.assertEqual(results, [
                'accept wait',
                'client start',
                'accept done',
                'handler start',
                ('handler', b'Msg1\n'),
                ('client', b'Msg1\n'),
                ('handler', b'Msg2\n'),
                ('client', b'Msg2\n'),
                'client close',
                'handler done'
                ])

    def test_udp_echo(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_DGRAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            results.append('recvfrom wait')
            data, addr = await sock.recvfrom(8192)
            results.append(('server', data))
            await sock.sendto(data, addr)
            await sock.close()
            results.append('server close')

        async def client(address):
            results.append('client start')
            sock = socket(AF_INET, SOCK_DGRAM)
            results.append('client send')
            await sock.sendto(b'Msg1', address)
            data, addr = await sock.recvfrom(8192)
            results.append(('client', data))
            await sock.close()
            results.append('client close')

        kernel.add_task(server(('',25000)))
        kernel.add_task(client(('localhost', 25000)))
        kernel.run()

        self.assertEqual(results, [
                'recvfrom wait',
                'client start',
                'client send',
                ('server', b'Msg1'),
                'server close',
                ('client', b'Msg1'),
                'client close'
                ])

    def test_accept_timeout(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_STREAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.listen(1)
            sock.settimeout(0.5)
            results.append('accept wait')
            try:
                client, addr = await sock.accept()
                results.append('not here')
            except TimeoutError:
                results.append('accept timeout')
            await sock.close()

        kernel.add_task(server(('',25000)))
        kernel.run()

        self.assertEqual(results, [
                'accept wait',
                'accept timeout'
                ])

    def test_accept_cancel(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_STREAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.listen(1)
            results.append('accept wait')
            try:
                client, addr = await sock.accept()
                results.append('not here')
            except CancelledError:
                results.append('accept cancel')
            await sock.close()

        async def canceller():
             task = await new_task(server(('',25000)))
             await sleep(0.5)
             await task.cancel()

        kernel.add_task(canceller())
        kernel.run()
        self.assertEqual(results, [
                'accept wait',
                'accept cancel'
                ])

    def test_recv_timeout(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_STREAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.listen(1)
            results.append('accept wait')
            client, addr = await sock.accept()
            results.append('recv wait')
            client.settimeout(0.5)
            try:
                data = await client.recv(8192)
                results.append('not here')
            except TimeoutError:
                results.append('recv timeout')
            await client.close()
            await sock.close()

        async def canceller():
             task = await new_task(server(('',25000)))
             sock = socket(AF_INET, SOCK_STREAM)
             results.append('client connect')
             await sock.connect(('localhost', 25000))
             await sleep(1.0)
             await sock.close()
             results.append('client done')

        kernel.add_task(canceller())
        kernel.run()

        self.assertEqual(results, [
                'accept wait',
                'client connect',
                'recv wait',
                'recv timeout',
                'client done'
                ])


    def test_recv_cancel(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_STREAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.listen(1)
            results.append('accept wait')
            client, addr = await sock.accept()
            results.append('recv wait')
            try:
                data = await client.recv(8192)
                results.append('not here')
            except CancelledError:
                results.append('recv cancel')
            await client.close()
            await sock.close()

        async def canceller():
             task = await new_task(server(('',25000)))
             sock = socket(AF_INET, SOCK_STREAM)
             results.append('client connect')
             await sock.connect(('localhost', 25000))
             await sleep(1.0)
             await task.cancel()
             await sock.close()
             results.append('client done')

        kernel.add_task(canceller())
        kernel.run()

        self.assertEqual(results, [
                'accept wait',
                'client connect',
                'recv wait',
                'recv cancel',
                'client done'
                ])

    def test_recvfrom_timeout(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_DGRAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            sock.settimeout(0.5)
            results.append('recvfrom wait')
            try:
                await sock.recvfrom(8192)
                results.append('not here')
            except TimeoutError:
                results.append('recvfrom timeout')
            await sock.close()

        async def canceller():
             await new_task(server(('',25000)))
             await sleep(1.0)
             results.append('client done')

        kernel.add_task(canceller())
        kernel.run()

        self.assertEqual(results, [
                'recvfrom wait',
                'recvfrom timeout',
                'client done'
                ])


    def test_recvfrom_cancel(self):
        kernel = get_kernel()
        results = []
        async def server(address):
            sock = socket(AF_INET, SOCK_DGRAM)
            sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
            sock.bind(address)
            results.append('recvfrom wait')
            try:
                await sock.recvfrom(8192)
                results.append('not here')
            except CancelledError:
                results.append('recvfrom cancel')
            await sock.close()

        async def canceller():
             task = await new_task(server(('',25000)))
             await sleep(1.0)
             await task.cancel()
             results.append('client done')

        kernel.add_task(canceller())
        kernel.run()

        self.assertEqual(results, [
                'recvfrom wait',
                'recvfrom cancel',
                'client done'
                ])

    def test_buffer_into(self):
        from array import array
        kernel = get_kernel()
        results = []
        async def sender(s1):
            a = array('i', range(1000000))
            await s1.sendall(a)
            
        async def receiver(s2):
            a = array('i', (0 for n in range(1000000)))
            view = memoryview(a).cast('b')
            total = 0
            while view:
                nrecv = await s2.recv_into(view)
                if not nrecv:
                    break
                total += nrecv
                view = view[nrecv:]

            results.append(a)

        s1, s2 = socketpair()
        kernel.add_task(sender(s1))
        kernel.add_task(receiver(s2))
        kernel.run()
        s1._socket.close()
        s2._socket.close()

        self.assertTrue(all(n==x for n,x in enumerate(results[0])))

if __name__ == '__main__':
    unittest.main()



# curio/test/workers.py

import unittest
import time

from ..import *

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n-1) + fib(n-2)

class TestWorkers(unittest.TestCase):
    def test_cpu(self):
        kernel = get_kernel()
        results = []

        async def spin(n):
            while n > 0:
                results.append(n)
                await sleep(0.1)
                n -= 1
                
        async def cpu_bound(n):
             r = await run_cpu_bound(fib, n)
             results.append(('fib', r))

        kernel.add_task(spin(10))
        kernel.add_task(cpu_bound(36))
        kernel.run()

        self.assertEqual(results, [
                10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
                ('fib', 14930352)
                ])

    def test_blocking(self):
        kernel = get_kernel()
        results = []

        async def spin(n):
            while n > 0:
                results.append(n)
                await sleep(0.1)
                n -= 1
                
        async def blocking(n):
             await run_blocking(time.sleep, n)
             results.append('sleep done')

        kernel.add_task(spin(10))
        kernel.add_task(blocking(2))
        kernel.run()

        self.assertEqual(results, [
                10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
                'sleep done',
                ])


    def test_blocking_timeout(self):
        kernel = get_kernel()
        results = []

        async def spin(n):
            while n > 0:
                results.append(n)
                await sleep(0.1)
                n -= 1
                
        async def blocking(n):
             try:
                 await run_blocking(time.sleep, n, timeout=0.55)
             except TimeoutError:
                 results.append('timeout')

        kernel.add_task(spin(10))
        kernel.add_task(blocking(2))
        kernel.run()

        self.assertEqual(results, [
                10, 9, 8, 7, 6, 5, 'timeout', 4, 3, 2, 1
                ])



if __name__ == '__main__':
    unittest.main()



# curio/sync.py
#
# Copyright (C) 2015
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# Implementation of common task synchronization primitives such as
# events, locks, semaphores, and condition variables. These primitives
# are only safe to use in the curio framework--they are not thread safe.

from .kernel import _wait_on_queue, _reschedule_tasks, kqueue

__all__ = ['Event', 'Lock', 'Semaphore', 'BoundedSemaphore', 'Condition' ]

class Event(object):
    __slots__ = ('_set', '_waiting')
    def __init__(self):
        self._set = False
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'set' if self._set else 'unset'
        return '<{} [{},waiters:{}]>'.format(res[1:-1], extra, len(self._waiting))

    def is_set(self):
        return self._set

    def clear(self):
        self._set = False

    async def wait(self, *, timeout=None):
        if self._set:
            return
        await _wait_on_queue(self._waiting, 'EVENT_WAIT', timeout)

    async def set(self):
        self._set = True
        await _reschedule_tasks(self._waiting, len(self._waiting))

class _LockBase(object):
    async def __aenter__(self):
        await self.acquire()
        return None

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()

class Lock(_LockBase):
    __slots__ = ('_acquired', '_waiting')
    def __init__(self):
        self._acquired = False
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return '<{} [{},waiters:{}]>'.format(res[1:-1], extra, len(self._waiting))

    async def acquire(self, *, timeout=None):
        if self._acquired:
            await _wait_on_queue(self._waiting, 'LOCK_ACQUIRE', timeout)
        self._acquired = True
        return True

    async def release(self):
        assert self._acquired, 'Lock not acquired'
        if self._waiting:
            await _reschedule_tasks(self._waiting, n=1)
        else:
            self._acquired = False

    def locked(self):
        return self._acquired

class Semaphore(_LockBase):
    __slots__ = ('_value', '_waiting')
    def __init__(self, value=1):
        self._value = value
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return '<{} [{},value:{},waiters:{}]>'.format(res[1:-1], extra, self._value, len(self._waiting))

    async def acquire(self, *, timeout=None):
        if self._value <= 0:
            await _wait_on_queue(self._waiting, 'SEMA_ACQUIRE', timeout)
        else:
            self._value -= 1
        return True

    async def release(self):
        if self._waiting:
            await _reschedule_tasks(self._waiting, n=1)
        else:
            self._value += 1
        
    def locked(self):
        return self._value == 0

class BoundedSemaphore(Semaphore):
    __slots__ = ('_bound_value',)
    def __init__(self, value=1):
        self._bound_value = value
        super().__init__(value)

    async def release(self):
        if self._value >= self._bound_value:
            raise ValueError('BoundedSemaphore released too many times')
        await super().release()

class Condition(_LockBase):
    __slots__ = ('_lock', '_waiting')
    def __init__(self, lock=None):
        if lock is None:
            self._lock = Lock()
        else:
            self._lock = lock
        self._waiting = kqueue()

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        return '<{} [{},waiters:{}]>'.format(res[1:-1], extra, len(self._waiting))

    def locked(self):
        return self._lock.locked()

    async def acquire(self, *, timeout=None):
        await self._lock.acquire(timeout=timeout)

    async def release(self):
        await self._lock.release()

    async def wait(self, *, timeout=None):
        if not self.locked():
            raise RuntimeError("Can't wait on unacquired lock")
        await self.release()
        try:
            await _wait_on_queue(self._waiting, 'COND_WAIT', timeout)
        finally:
            await self.acquire()

    async def wait_for(self, predicate, *, timeout=None):
        while True:
            result = predicate()
            if result:
                return result
            await self.wait(timeout=timeout)

    async def notify(self, n=1):
        if not self.locked():
            raise RuntimeError("Can't notify on unacquired lock")
        await _reschedule_tasks(self._waiting, n=n)

    async def notify_all(self):
        await self.notify(len(self._waiting))




# curio/subprocess.py
#
# Copyright (C) 2015
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# Curio clone of the subprocess module.  

from .kernel import  new_task, sleep
from .io import Stream
import subprocess

__all__ = [ 'run', 'Popen', 'CompletedProcess', 'CalledProcessError', 'TimeoutExpired', 'SubprocessError',
            'check_output', 'PIPE', 'STDOUT', 'DEVNULL' ]

from subprocess import (
    CompletedProcess,
    SubprocessError, 
    CalledProcessError,
    TimeoutExpired,
    PIPE,
    STDOUT,
    DEVNULL,
    )

class Popen(object):
    '''
    Curio wrapper around the Popen class from the subprocess module. All of the
    methods from subprocess.Popen should be available, but the associated file
    objects for stdin, stdout, stderr have been replaced by async versions.
    Certain blocking operations (e.g., wait() and communicate()) have been
    replaced by async compatible implementations.
    '''
    def __init__(self, args, **kwargs):
        if 'universal_newlines' in kwargs:
            raise RuntimeError('universal_newlines argument not supported')

        self._popen = subprocess.Popen(args, **kwargs)
        if self._popen.stdin:
            self.stdin = Stream(self._popen.stdin)
        if self._popen.stdout:
            self.stdout = Stream(self._popen.stdout)
        if self._popen.stderr:
            self.stderr = Stream(self._popen.stderr)

    def __getattr__(self, name):
        return getattr(self._popen, name)

    async def wait(self, timeout=None):
        async def waiter():
            while True:
                retcode = self._popen.poll()
                if retcode is not None:
                    return retcode
                await sleep(0.0005)
        task = await new_task(waiter())
        try:
            return await task.join(timeout=timeout)
        except TimeoutError:
            await task.cancel()
            raise TimeoutExpired(self.args, timeout) from None

    async def communicate(self, input=b'', timeout=None):
        if input:
            assert self.stdin
            async def writer():
                await self.stdin.write(input)
                await self.stdin.close()
            stdin_task = await writer()
        else:
            stdin_task = None

        stdout_task = await new_task(self.stdout.readall()) if self.stdout else None
        stderr_task = await new_task(self.stderr.readall()) if self.stderr else None

        # Collect the output from the workers
        try:
            if stdin_task:
                await stdin_task.join(timeout=timeout)
            stdout = await stdout_task.join(timeout=timeout) if stdout_task else b''
            stderr = await stderr_task.join(timeout=timeout) if stderr_task else b''
            return (stdout, stderr)
        except TimeoutError:
            await stdout_task.cancel()
            await stderr_task.cancel()
            if stdin_task:
                await stdin_task.cancel()
            raise

    def __enter__(self):
        raise RuntimeError('Use async-with')

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        if self.stdout:
            await self.stdout.close()

        if self.stderr:
            await self.stderr.close()

        if self.stdin:
            await self.stdin.close()

        # Wait for the process to terminate
        await self.wait()

async def run(args, *, stdin=None, input=None, stdout=None, stderr=None, shell=False, timeout=None, check=False):
    '''
    Curio-compatible version of subprocess.run()
    '''
    if input:
        stdin = subprocess.PIPE
    else:
        stdin = None

    async with Popen(args, stdin=stdin, stdout=stdout, stderr=stderr, shell=shell) as process:
        try:
            stdout, stderr = await process.communicate(input, timeout)
        except TimeoutError:
            process.kill()
            stdout, stderr = await process.communicate()
            raise TimeoutExpired(process.args, timeout, output=stdout, stderr=stderr)
        except:
            process.kill()
            raise

    retcode = process.poll()
    if check and retcode:
        raise CalledProcessError(retcode, process.args,
                                 output=stdout, stderr=stderr)
    return CompletedProcess(process.args, retcode, stdout, stderr)

async def check_output(args, *, stdin=None, stderr=None, shell=False, timeout=None):
     out = await run(args, stdout=PIPE, stdin=stdin, stderr=stderr, shell=shell, timeout=timeout, check=True)
     return out.stdout



# Echo server with cancellation and signal handling

import signal
from curio import Kernel, new_task, SignalSet, CancelledError, run_server

async def echo_client(client, addr):
    print('Connection from', addr)
    try:
        while True:
            data = await client.recv(1000)
            if not data:
                break
            await client.sendall(data)
        print('Connection closed')
    except CancelledError:
        await client.sendall(b'Server going down\n')
    
async def main(host, port):
    while True:
        async with SignalSet(signal.SIGHUP) as sigset:
            print('Starting the server')
            serv_task = await new_task(run_server(host, port, echo_client))
            await sigset.wait()
            print('Server shutting down')
            await serv_task.cancel_children()
            await serv_task.cancel()

if __name__ == '__main__':
    kernel = Kernel(with_monitor=True)
    kernel.run(main('', 25000))



# curio/test/socket.py

import unittest
from ..import *
from ..socket import *

class TestSocketServer(unittest.TestCase):
    def test_tcp_echo_baserequest(self):
        kernel = get_kernel()
        results = []

        async def handler(client, addr):
            results.append('handler start')
            while True:
                results.append('recv wait')
                data = await client.recv(100)
                if not data:
                    break
                results.append(('handler', data))
                await client.sendall(data)
            results.append('handler done')

        async def client(address, serv):
            results.append('client start')
            sock = socket(AF_INET, SOCK_STREAM)
            await sock.connect(address)
            await sock.send(b'Msg1')
            await sleep(0.1)
            resp = await sock.recv(100)
            results.append(('client', resp))
            await sock.send(b'Msg2')
            await sleep(0.1)
            resp = await sock.recv(100)
            results.append(('client', resp))
            results.append('client close')
            await sock.close()
            await serv.cancel()

        serv = kernel.add_task(run_server('',25000,handler))
        kernel.add_task(client(('localhost', 25000), serv))
        kernel.run()

        self.assertEqual(results, [
                'client start',
                'handler start',
                'recv wait',
                ('handler', b'Msg1'),
                'recv wait',
                ('client', b'Msg1'),
                ('handler', b'Msg2'),
                'recv wait',
                ('client', b'Msg2'),
                'client close',
                'handler done'
                ])

if __name__ == '__main__':
    unittest.main()



# ssl_echo
#
# An example of a simple SSL server.  To test, connect via browser

import curio
from curio import ssl
import time

KEYFILE = "ssl_test_rsa"    # Private key
CERTFILE = "ssl_test.crt"   # Certificate (self-signed)

async def handler(client, addr):
    reader, writer = client.make_streams()
    async with reader, writer:
        async for line in reader:
            line = line.strip()
            if not line:
                break
            print(line)

        await writer.write(
b'''HTTP/1.0 200 OK\r
Content-type: text/plain\r
\r
If you're seeing this, it probably worked. Yay!
''')
        await writer.write(time.asctime().encode('ascii'))

if __name__ == '__main__':
    kernel = curio.Kernel()
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(certfile=CERTFILE, keyfile=KEYFILE)
    kernel.run(curio.run_server('', 10000, handler, ssl=ssl_context))



# echoserv.py
#
# Echo server using streams

from curio import Kernel, new_task, run_server

async def echo_client(client, addr):
    print('Connection from', addr)
    reader, writer = client.make_streams()
    async with reader, writer:
        async for line in reader:
            await writer.write(line)
    print('Connection closed')

if __name__ == '__main__':
    kernel = Kernel()
    kernel.run(run_server('', 25000, echo_client))



# curio/test/kernel.py

import unittest
import time
from ..import *

# ---- Basic Kernel Functionality

class TestKernel(unittest.TestCase):
    def test_hello(self):
        kernel = get_kernel()
        results = []
        async def hello():
              results.append('hello')

        kernel.add_task(hello())
        kernel.run()
        self.assertEqual(results, [
                'hello'
                ])

    def test_sleep(self):
        kernel = get_kernel()
        results = []

        async def main():
              results.append('start')
              await sleep(0.5)
              results.append('end')

        kernel.add_task(main())
        start = time.time()
        kernel.run()
        end = time.time()
        self.assertEqual(results, [
                'start',
                'end',
                ])
        self.assertTrue((end-start) > 0.5)

    def test_sleep_cancel(self):
        kernel = get_kernel()
        results = []

        async def sleeper():
            results.append('start')
            try:
                await sleep(1)
                results.append('not here')
            except CancelledError:
                results.append('cancelled')

        async def main():
            task = await new_task(sleeper())
            await sleep(0.5)
            await task.cancel()

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'start',
                'cancelled',
                ])

    def test_task_join(self):
        kernel = get_kernel()
        results = []

        async def child():
            results.append('start')
            await sleep(0.5)
            results.append('end')
            return 37

        async def main():
            task = await new_task(child())
            await sleep(0.1)
            results.append('joining')
            r = await task.join()
            results.append(r)

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'start',
                'joining',
                'end',
                37
                ])

    def test_task_join_error(self):
        kernel = get_kernel()
        results = []

        async def child():
            results.append('start')
            int('bad')

        async def main():
            task = await new_task(child())
            await sleep(0.1)
            results.append('joining')
            try:
                r = await task.join()
                results.append(r)
            except TaskError as e:
                results.append('task fail')
                results.append(type(e))
                results.append(type(e.__cause__))

        kernel.add_task(main())
        kernel.run(log_errors=False)
        self.assertEqual(results, [
                'start',
                'joining',
                'task fail',
                TaskError,
                ValueError,
                ])

    def test_task_cancel(self):
        kernel = get_kernel()
        results = []

        async def child():
            results.append('start')
            try:
                await sleep(0.5)
                results.append('end')
            except CancelledError:
                results.append('cancelled')

        async def main():
            task = await new_task(child())
            results.append('cancel start')
            await sleep(0.1)
            results.append('cancelling')
            await task.cancel()
            results.append('done')

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'start',
                'cancel start',
                'cancelling',
                'cancelled',
                'done',
                ])

    def test_task_child_cancel(self):
        kernel = get_kernel()
        results = []

        async def child():
            results.append('start')
            try:
                await sleep(0.5)
                results.append('end')
            except CancelledError:
                results.append('child cancelled')

        async def parent():
            try:
                 child_task = await new_task(child())
                 await sleep(0.5)
                 results.append('end parent')
            except CancelledError:
                await child_task.cancel()
                results.append('parent cancelled')
            
        async def grandparent():
            try:
                parent_task = await new_task(parent())
                await sleep(0.5)
                results.append('end grandparent')
            except CancelledError:
                await parent_task.cancel()
                results.append('grandparent cancelled')

        async def main():
            task = await new_task(grandparent())
            await sleep(0.1)
            results.append('cancel start')
            await sleep(0.1)
            results.append('cancelling')
            await task.cancel()
            results.append('done')

        kernel.add_task(main())
        kernel.run()

        self.assertEqual(results, [
                'start',
                'cancel start',
                'cancelling',
                'child cancelled',
                'parent cancelled',
                'grandparent cancelled',
                'done',
                ])

    def test_task_ready_cancel(self):
        # This tests a tricky corner case of a task cancelling another task that's also 
        # on the ready queue.

        kernel = get_kernel()
        results = []

        async def child():
            try:
                results.append('child sleep')
                await sleep(1.0)
                results.append('child slept')
                await sleep(1.0)
                results.append('should not see this')
            except CancelledError:
                results.append('child cancelled')

        async def parent():
            task = await new_task(child())
            results.append('parent sleep')
            await sleep(0.5)
            results.append('cancel start')
            await task.cancel()
            results.append('cancel done')
            
        async def main():
            task = await new_task(parent())
            await sleep(0.1)
            time.sleep(1)      # Forced block of the event loop. Both tasks should awake when we come back
            await sleep(0.1)

        kernel.add_task(main())
        kernel.run()

        self.assertEqual(results, [
                'child sleep',
                'parent sleep',
                'cancel start',
                'child slept',
                'child cancelled',
                'cancel done'
                ])

class TestSignal(unittest.TestCase):
    def test_task_signal(self):
        import signal, os
        kernel = get_kernel()
        results = []

        async def child():
            async with SignalSet(signal.SIGUSR1, signal.SIGUSR2) as sig:
                signo = await sig.wait()
                results.append(signo)
                signo = await sig.wait()
                results.append(signo)

        async def main():
            task = await new_task(child())
            await sleep(0.1)
            results.append('sending USR1')
            os.kill(os.getpid(), signal.SIGUSR1)
            await sleep(0.1)
            results.append('sending USR2')
            os.kill(os.getpid(), signal.SIGUSR2)
            await sleep(0.1)
            results.append('done')

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'sending USR1',
                signal.SIGUSR1,
                'sending USR2',
                signal.SIGUSR2,
                'done',
                ])


    def test_task_signal_waitone(self):
        import signal, os
        kernel = get_kernel()
        results = []

        async def child():
            sig = SignalSet(signal.SIGUSR1, signal.SIGUSR2)
            signo = await sig.wait()
            results.append(signo)
            signo = await sig.wait()
            results.append(signo)

        async def main():
            task = await new_task(child())
            await sleep(0.1)
            results.append('sending USR1')
            os.kill(os.getpid(), signal.SIGUSR1)
            await sleep(0.1)
            results.append('sending USR2')
            os.kill(os.getpid(), signal.SIGUSR2)
            await sleep(0.1)
            results.append('done')

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'sending USR1',
                signal.SIGUSR1,
                'sending USR2',
                signal.SIGUSR2,
                'done',
                ])


    def test_task_signal_ignore(self):
        import signal, os
        kernel = get_kernel()
        results = []

        async def child():
            sig = SignalSet(signal.SIGUSR1, signal.SIGUSR2)
            async with sig:
                 signo = await sig.wait()
                 results.append(signo)
                 with sig.ignore():
                     await sleep(1)
                 results.append('here')

        async def main():
            task = await new_task(child())
            await sleep(0.1)
            results.append('sending USR1')
            os.kill(os.getpid(), signal.SIGUSR1)
            await sleep(0.5)
            results.append('sending USR1')
            os.kill(os.getpid(), signal.SIGUSR1)
            await sleep(0.1)
            await task.join()
            results.append('done')

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'sending USR1',
                signal.SIGUSR1,
                'sending USR1',
                'here',
                'done',
                ])

    def test_task_signal_timeout(self):
        import signal, os
        kernel = get_kernel()
        results = []

        async def child():
            async with SignalSet(signal.SIGUSR1, signal.SIGUSR2) as sig:
                try:
                    signo = await sig.wait(timeout=0.5)
                    results.append(signo)
                except TimeoutError:
                    results.append('timeout')

        async def main():
            task = await new_task(child())
            await task.join()
            results.append('done')

        kernel.add_task(main())
        kernel.run()
        self.assertEqual(results, [
                'timeout',
                'done',
                ])

if __name__ == '__main__':
    unittest.main()



# curio/workers.py
#
# Copyright (C) 2015
# David Beazley (Dabeaz LLC), http://www.dabeaz.com
# All rights reserved.
#
# Functions for performing work outside of curio.  This includes running functions
# in thread pools, process pools, etc.

from .kernel import _future_wait, new_task, sleep
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

__all__ = [ 'run_in_executor', 'run_blocking', 'run_cpu_bound', 'set_blocking_executor', 'set_cpu_executor' ]

_MAX_BLOCKING_THREADS = 16

_default_blocking_executor = None
_default_cpu_executor = None

def set_blocking_executor(exc):
    '''
    Set the executor used to execute code with the run_blocking() function.
    '''
    global _default_blocking_executor
    _default_blocking_executor = exc

def set_cpu_executor(exc):
    '''
    Set the executor used to execute code with the run_cpu_bound() function.
    '''
    global _default_cpu_executor
    _default_cpu_executor = exc

async def run_in_executor(exc, callable, *args, timeout=None):
    '''
    Run a callable in an executor (e.g., concurrent.futures.ThreadPoolExecutor) and return the result.
    '''
    future = exc.submit(callable, *args)
    await _future_wait(future, timeout=timeout)
    return future.result()

async def run_blocking(callable, *args, timeout=None):
    if _default_blocking_executor is None:
        set_blocking_executor(ThreadPoolExecutor(_MAX_BLOCKING_THREADS))
        
    return await run_in_executor(_default_blocking_executor, callable, *args, timeout=timeout)

async def run_cpu_bound(callable, *args, timeout=None):
    if _default_cpu_executor is None:
        set_cpu_executor(ProcessPoolExecutor())
    return await run_in_executor(_default_cpu_executor, callable, *args, timeout=timeout)

    



# An example of a server involving a CPU-intensive task.  We'll farm the 
# CPU-intensive work out to a separate process.

from curio import Kernel, new_task, run_cpu_bound, run_server

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n-1) + fib(n-2)

async def fib_handler(client, addr):
    print('Connection from', addr)
    rfile, wfile = client.make_streams()
    async with rfile, wfile:
        async for line in rfile:
            try:
                n = int(line)
                result = await run_cpu_bound(fib, n)
                resp = str(result) + '\n'
                await wfile.write(resp.encode('ascii'))
            except ValueError:
                await wfile.write(b'Bad input\n')
    print('Connection closed')

if __name__ == '__main__':
    kernel = Kernel()
    kernel.run(run_server('', 25000, fib_handler))



    
